#
#  Copyright 1996-2014 Jim Pirzyk / pirzyk.org
#  All rights reserved.
#
#  Security::sudo - sudo control module
#
package Security::sudo;

use Security::util::die;
use Security::util::print;
use Security::util::read_config;
use Security::db;
use Security::util::stat;
use Security::util::users;
use Security::util::groups;
use Security::util::uniq;
use Exporter;
use POSIX;
use strict;
use warnings "all";

our @ISA = qw(Exporter);
our @EXPORT_OK = qw( init check print );

our (%sudoworld, %sudogroup, %sudoowner, %sudomissing);

sub init {
	print ( STDERR  "In Security::sudo::init()\n") if ( $main::opts{'d'} );

	&read_config('nosudoowner');
}

# Create a canonical list of all users who are a member of the $gid group
sub _get_group_members () {
	my ($gid) = @_;
	my (%grps);

	# Start with the list of members from the /etc/group file
	map { $grps{$_} = 1; } &Security::util::groups::get('gid',$gid,'members');

	# Now add all the accounts that have this group as the primary gid
	foreach my $name ( &Security::util::users::get('name')) {
		my ($g) = &Security::util::users::get('name', $name, 'gid');
		$grps{$name} = 1
			if ($g eq $gid);
	}

	return keys %grps;
}

sub check {
	print ( STDERR  "In Security::sudo::check()\n") if ( $main::opts{'d'} );

	my ($FP, $currsum, $cmds, $cmd, $who);
	open ($FP, $main::sudofile ) || &security_die ("Could not read sudoers file ($main::sudofile)\n");
	while (<$FP>) {
		chomp;
		# The sample sudo line:
		# dgb    boulder = (operator) /bin/ls, /bin/kill, /usr/bin/lprm
		# $1 == from who
		# $2 == host
		# $3 == who as
		# $4 ?= TagSpec:
		# $4 or $5 == commands to allow
		if ( ! /^Default/o && /(\S+)\s+(\S+)\s*=\s*\((\S+)\)(\s*\S+:)?\s*(.*)$/ ) {
			$who = $3;
			my @cmds = ( defined $5
				? split (/,\s*/, $5)
				: split (/,\s*/, $4));

			# Now find the command.
			foreach $cmd ( @cmds ) {
				# Now expand any wildcarding in the $cmd
				if ( defined $cmd && length $cmd && defined $who && length $who ) {
					foreach my $cmd2 ( glob $cmd ) {
						push @{$cmds->{$cmd2}->{who}}, $who
							if ( $cmd2 ne 'ALL' );
					}
				}
			}
		}
	}
	close $FP;

	# Clean up the who list to have only unique names
	foreach $cmd ( keys %$cmds ) {
		@{$cmds->{$cmd}->{who}} = &Security::util::uniq::uniq(@{$cmds->{$cmd}->{who}});
	}

	# Now get mode/uid/gid info for each program 
	# and check who can write to the file.
	foreach $cmd ( keys %$cmds ) {
 		if ( ! -f $cmd ) {
			$sudomissing{$cmd} = 1;		
 		} else {
			&Security::db::verify($cmd);

			my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$curbytes) = 
				&Security::util::stat::get_stat($cmd, 'stat');

			# If the file is world writeable, then flag it
			$sudoworld{$cmd} = 1
				if ( $mode & &S_IWOTH() );

			# Now check if it is group writeable, and who is in the group
			if ( $mode & &S_IWGRP() ) {
				my @list = &_get_group_members ($gid);
				($who) = &Security::util::users::get('uid', $uid, 'name');

				# If the owner is NOT in the list of group members, then
				# complain about this.
				if (!grep (/^\Q${who}\E$/, @list)) {
					$sudogroup{$cmd} = 1;
				}
			}

			# Lastly, check to make sure if the file is user writeable
			# and who owns the file (exclude files owned by root or
			# bin (uid=$binUID).  This can be ignore if nosudoowner is
			# set in the config file.
			if (( !exists $main::config{'nosudoowner'} || ! defined $main::config{'nosudoowner'} ||  $main::config{'nosudoowner'} ne 'yes' ) && $mode & &S_IWUSR() && $uid != 0 && $uid != $main::binuid ) {
				($who) = &Security::util::users::get('uid', $uid, 'name');
				if (! grep (/^\Q${who}\E$/, @{$cmds->{$cmd}->{who}})) {
					$sudoowner{$cmd} = 1;
				}
			}
		}
	}
}

sub print {
	print ( STDERR  "In Security::sudo::print()\n") if ( $main::opts{'d'} );

	&print_data ("\n\nchecking sudo missing commands:\n", %sudomissing);
	&print_data ("\n\nchecking for world writable sudo commands:\n", %sudoworld);
	&print_data ("\n\nchecking for group writable sudo commands:\n", %sudogroup);
	&print_data ("\n\nchecking for owner mismatch sudo commands:\n", %sudoowner);
}

1;

__END__

=head1 NAME

Security::sudo - sudo control module

=head1 SYNOPSIS

Security::sudo::init()

Security::sudo::check()

Security::sudo::print()

=head1 DESCRIPTION

Check the owner, group and mode of all commands configured in the
F<sudoers> file.  The command should be owned by either C<root>, (C<bin>
on Solaris) or the account the command is going to be run as (usually this
is C<root> as well).

=head1 API

This module implements the default API:

=over 4

=item C<init()>

Calls C<read_config (nosudoowner)> to see if we do not want to test file
ownership.

=item C<check()>

Parses the F<sudoers> file, looking for all commands that are configured
and who the command is going to be run as.

=item C<print()>

Calls C<print_data ()> to print out the list of commands in the F<sudoers>
file but do not exist on the system.

Calls C<print_data ()> to print out the list of commands in the F<sudoers>
file that are world writable.

Calls C<print_data ()> to print out the list of commands in the F<sudoers>
file that are group writable.

Calls C<print_data ()> to print out the list of commands in the F<sudoers>
file that are not owned by the account they are being run as.

=back

=head1 CONFIGURATION OPTIONS

The script will call C<read_config> on the following config file options:

=over 4

=item C<nosudoowner>

Turn off the test for ownership of sudo commands.

=back

=head1 HISTORY

the C<nosudoowner> option was suggested by a co-worker at UIUC, for scripts
that were not owned by who they were being executed as.

=head1 SEE ALSO

L<Security::db>

L<Security::util::groups>

L<Security::util::stat>

L<Security::util::uniq>

L<Security::util::users>

L<sudo(8)>

=cut

